# Spring Security 权限控制

---

![](https://gitee.com/veal98/images/raw/master/img/20210130143717.png)

导入 依赖：

```xml
<!--Spring Security-->
<dependency>
    <groupId>org.springframework.boot</groupId>
    <artifactId>spring-boot-starter-security</artifactId>
</dependency>
```

## 授权配置

![](https://gitee.com/veal98/images/raw/master/img/20210130152216.png)

管理员能够访问的特殊路径后续再配

```java
@Configuration
public class SecurityConfig extends WebSecurityConfigurerAdapter implements CommunityConstant {

    /**
     * 静态资源
     * @param web
     * @throws Exception
     */
    @Override
    public void configure(WebSecurity web) throws Exception {
        web.ignoring().antMatchers("/resources/**");
    }

    // 认证环节我们使用自己的代码 LoginController，绕过 Spring Security 的

    /**
     * 授权
     * @param http
     * @throws Exception
     */
    @Override
    protected void configure(HttpSecurity http) throws Exception {
        http.authorizeRequests()
                .antMatchers(
                    "/user/setting",
                        "/user/upload",
                        "/discuss/add",
                        "/comment/add",
                        "/letter/**",
                        "/notice/**",
                        "/like",
                        "/follow",
                        "/unfollow"
                )
                .hasAnyAuthority(
                        AUTHORITY_USER,
                        AUTHORITY_ADMIN,
                        AUTHORITY_MODERATOR
                )
                .anyRequest().permitAll();

        // 权限不够时的处理
        http.exceptionHandling()
                // 1. 未登录时的处理
                .authenticationEntryPoint(new AuthenticationEntryPoint() {
                    @Override
                    public void commence(HttpServletRequest request, HttpServletResponse response, AuthenticationException e) throws IOException, ServletException {
                        String xRequestedWith = request.getHeader("x-requested-with");
                        if ("XMLHttpRequest".equals(xRequestedWith)) {
                            // 异步请求
                            response.setContentType("application/plain;charset=utf-8");
                            PrintWriter writer = response.getWriter();
                            writer.write(CommunityUtil.getJSONString(403, "你还没有登录"));
                        }
                        else {
                            // 普通请求
                            response.sendRedirect(request.getContextPath() + "/login");
                        }
                    }
                })
                // 2. 权限不够时的处理
                .accessDeniedHandler(new AccessDeniedHandler() {
                    @Override
                    public void handle(HttpServletRequest request, HttpServletResponse response, AccessDeniedException e) throws IOException, ServletException {
                        String xRequestedWith = request.getHeader("x-requested-with");
                        if ("XMLHttpRequest".equals(xRequestedWith)) {
                            // 异步请求
                            response.setContentType("application/plain;charset=utf-8");
                            PrintWriter writer = response.getWriter();
                            writer.write(CommunityUtil.getJSONString(403, "你没有访问该功能的权限"));
                        }
                        else {
                            // 普通请求
                            response.sendRedirect(request.getContextPath() + "/denied");
                        }
                    }
                });

        // Security 底层会默认拦截 /logout 请求，进行退出处理
        // 此处赋予它一个根本不存在的退出路径，使得程序能够执行到我们自己编写的退出代码
        http.logout().logoutUrl("/securitylogout");
    }
}
```

Security 底层会默认拦截 /logout 请求，进行退出处理。此处赋予它一个根本不存在的退出路径 /securitylogout（善意的谎言），使得程序能够执行到我们自己编写的退出代码

## 认证方案

认证环节我们之前已经写好了，就不用 Security 的了。即绕过 Spring Security 的认证流程，使用我们自己编写的代码。不在下面 Security 提供的认证方法中写：

```java
@Override
protected void configure(AuthenticationManagerBuilder auth) throws Exception {
    super.configure(auth);
}
```

**既然我们需要自己做认证，也就是将认证的结果（用户、密码、权限）存入 SecurityContext**，便于 Security 的授权：

我们之前利用拦截器  LoginTicketInterceptor 在所有 Controller 执行之前获取了用户信息，所以在此处做如下修改：

![](https://gitee.com/veal98/images/raw/master/img/20210130152730.png)

其中，需要在 UserService 中添加一个获取用户权限的方法：

```java
/**
 * 获取某个用户的权限
 * @param userId
 * @return
 */
public Collection<? extends GrantedAuthority> getAuthorities(int userId) {
    User user = this.findUserById(userId);
    List<GrantedAuthority> list = new ArrayList<>();
    list.add(new GrantedAuthority() {
        @Override
        public String getAuthority() {
            switch (user.getType()) {
                case 1:
                    return AUTHORITY_ADMIN;
                case 2:
                    return AUTHORITY_MODERATOR;
                default:
                    return AUTHORITY_USER;
            }
        }
    });
    return list;
}
```

别忘了请求结束后清除相关内容：

![](https://gitee.com/veal98/images/raw/master/img/20210130152843.png)

![](https://gitee.com/veal98/images/raw/master/img/20210130152910.png)

## CSRF 配置

禁用 csrf 的 token 检查

![](https://gitee.com/veal98/images/raw/master/img/20210130201754.png)